/*	findslides.c: locate slides in large high-res pnm file and
	break out as smaller files.
	Expected size of a slide is about 1.33 by 0.91 inches
	T. Wallace, 1/29/04

	(C) Copyright 2004 Timothy P. Wallace
	All Rights Reserved
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <ctype.h>
#include "debug.h"

#define PROG "findslides"
#define TYP_WID 1.33
#define TYP_HT 0.91
/* Following is distance from the center of the slide to certain blackness */
#define HALF_BND_SPACE 0.86
#define Xtop 0
#define Ytop 1
#define Xbot 2
#define Ybot 3
#define Xcent 4
#define Ycent 5

#define HORIZ 0
#define VERT 1

void syntax( int exitval );
void *alloc( int nbytes );
FILE *ftry( char *name, char *mode);
void fatal( char *msg, int level);
void process(FILE *fp, int resolution,int nextnum);
char *getlin(FILE *fp, int len, char *buf);
int find_non_black(unsigned char *image,int xdim,int ydim,int *bbox,int nrows,
	double exp_wid, double exp_ht, double slope);
int find_step_increase(double *energy,int start,int fini,int nrows,double exp_size,
	double slope);
void output_slide(int num,unsigned char *image,int xsize,int *bbox,int resolution);
void process_1_slide(unsigned char *image,int xdim,int ydim,int *bbox, int resolution,
	int slidenum);
int size_check(int *bbox, int resolution, double *area, int slidenum,
	int orientation);

main(int argc,char **argv)
{
  extern char *optarg;
  extern int optind;
  int ch;
  int resolution;
  FILE *fp;

  int nextnum = 1;
	
  resolution = 1600;
  while ((ch=getopt(argc,argv,"d:Hhn:r:"))!=EOF) {
    switch (ch) {
	case 'n':
	  nextnum = atoi(optarg);
	  break;
	case 'r':
	  resolution = atoi(optarg);
	  break;
#ifdef DEBUG
	case 'd':
	  debug = atoi(optarg);
	  break;
#endif
	case 'H':
	case 'h':
	case '?':
		syntax(1);
    }
  }
  if (optind < argc)
    fp = ftry(argv[optind],"r");
  else fp = stdin;
  process(fp,resolution,nextnum);
  exit(0);
}

void syntax(int exitval)
{
	fprintf(stderr,"Syntax: %s [",PROG);
	fprintf(stderr," -n first_slide_num (default 0001)");
	fprintf(stderr,"\n\t");
	fprintf(stderr," -r resolution_dpi (default 1600)");
	fprintf(stderr," -h (thismsg)");
#ifdef DEBUG
	fprintf(stderr," -ddebugflag");
#endif
	fprintf(stderr," file]\n");
	exit(exitval);
}

void *alloc(int nbytes)
{	/* like malloc, but prints msg and exits on error */
	void *malloc();
	char *ptr;
	if ((ptr = malloc((size_t)nbytes)) == NULL) {
	  fprintf(stderr,"%s: can't find %d free bytes!\n",
			PROG,nbytes);
		exit(2);
	}
	return ptr;
}

FILE *ftry(char *name,char *mode)
{
	FILE *fptr, *fopen();

	if ((fptr = fopen(name,mode)) == NULL) {
		fprintf(stderr,"%s: can't open %s (mode %s).\n",PROG,
			name,mode);
		exit(2);
	}
	TRACE(5,(DF,"ftry: returning fp %x on file %s (mode %s)\n",fptr,name,mode));
	return fptr;
}

void fatal(char *msg,int level)
{
	fprintf(stderr,"%s: %s\n",PROG,msg);
	exit(level);
}

void process(FILE *fp, int resolution, int nextnum)
{
  char buf[512];
  int xsize, ysize, depth, nbytes;
  unsigned char *image;
  int nread;
  int bbox[6]; /* x1,y1, x2,y2, xcent,ycent of bounding box */
#ifdef DEBUG
  int i;
#endif

  getlin(fp,512,buf);
  if (toupper(buf[0]) != 'P') {
    fprintf(stderr,"Error: file type doesn't start with 'P'\n");
    return;
  }
  getlin(fp,512,buf);
  if (sscanf(buf,"%d %d",&xsize,&ysize) != 2) {
    fprintf(stderr,"Error on image dimensions decode from %s\n",buf);
    return;
  }
  getlin(fp,512,buf);
  if (sscanf(buf,"%d",&depth) != 1) {
    fprintf(stderr,"Error on image depth decode from %s\n",buf);
    return;
  }
  TRACE(1,(DF,"process says read image header: %d %d %d\n",xsize,ysize,depth));
/* make sure we have enough data for the resolution specified: */
  if ( (((double)xsize / resolution) < 1.8)  ||
	(((double)ysize / resolution) < 8.8) ) {
    fprintf(stderr,"Image too small for resolution %d dpi\n",resolution);
    exit(0);
  }
  nbytes = xsize * ysize * 3;
  image = (unsigned char *)alloc(nbytes);
  nread = fread(image,1,nbytes,fp);
  TRACE(1,(DF,"process says read %d image bytes\n",nread));
#ifdef DEBUG
  for (i=0; i<16; i++) {
    TRACE(20,(DF,"%d ",image[i]));
  }
  TRACE(20,(DF,"\n"));
#endif
  bbox[Xcent] = 0.925 * resolution;

  bbox[Ycent] = 1.05 * resolution;
  process_1_slide(image,xsize,ysize,bbox,resolution,nextnum++);

  bbox[Ycent] = 3.33 * resolution;
  process_1_slide(image,xsize,ysize,bbox,resolution,nextnum++);

  bbox[Ycent] = 5.61 * resolution;
  process_1_slide(image,xsize,ysize,bbox,resolution,nextnum++);

  bbox[Ycent] = 7.89 * resolution;
  process_1_slide(image,xsize,ysize,bbox,resolution,nextnum++);
}

void process_1_slide(unsigned char *image,int xdim,int ydim,int *bbox, int resolution,
	int slidenum)
{
  int nrows; /* # rows(cols) used to estimate ``black'' */
  double exp_wid, exp_ht; /* expected image size in pixels */
  /* weighting function slope; higher forces detection closer to expected edge */
  double slope;
  double area, bestarea, best_orient;
  int bestbbox[4];
  int tmpbbox[4];
  int orientation = HORIZ; /* true if slide is ordinary horizontal one */
  nrows = 0.0125 * resolution; /* 5 at 400 dpi */
  bbox[Xtop] = bbox[Xcent] - HALF_BND_SPACE * resolution;
  bbox[Xbot] = bbox[Xcent] + HALF_BND_SPACE * resolution;
  bbox[Ytop] = bbox[Ycent] - HALF_BND_SPACE * resolution;
  bbox[Ybot] = bbox[Ycent] + HALF_BND_SPACE * resolution;
  exp_wid = TYP_WID * resolution;
  exp_ht = TYP_HT * resolution;
  bestarea = 0;
  TRACE(1,(DF,"process_1_slide: trying horizontal orientation\n"));
  for (slope=1.0; slope < 5.1; slope += 0.5) {
    memcpy(tmpbbox,bbox,sizeof(int)*4);
    if (!find_non_black(image,xdim,ydim,tmpbbox,nrows,exp_wid,exp_ht,slope)) {
	fprintf(stderr,"Missing slide %d, skipping\n",slidenum);
	return;
    }
    if (size_check(tmpbbox,resolution,&area,0,orientation)) {
	output_slide(slidenum,image,xdim,tmpbbox,resolution);
	return;
    } else if (area > bestarea) {
	  bestarea = area;
	  memcpy(bestbbox,tmpbbox,sizeof(int)*4);
	  best_orient = HORIZ;
    }
  }
  orientation = VERT; /* try vertical slide detection */
  TRACE(1,(DF,"process_1_slide: trying vertical orientation\n"));
  for (slope=1.0; slope < 5.1; slope += 0.5) {
    memcpy(tmpbbox,bbox,sizeof(int)*4);
    if (!find_non_black(image,xdim,ydim,tmpbbox,nrows,exp_ht,exp_wid,slope)) {
	fprintf(stderr,"Missing slide %d, skipping\n",slidenum);
	return;
    }
    if (size_check(tmpbbox,resolution,&area,0,orientation)) {
	output_slide(slidenum,image,xdim,tmpbbox,resolution);
	return;
    } else if (area > bestarea) {
	  bestarea = area;
	  memcpy(bestbbox,tmpbbox,sizeof(int)*4);
	  best_orient = VERT;
    }
  }
  /* Nothing quite fit our minimum size; output largest slide found */
  size_check(tmpbbox,resolution,&area,slidenum,best_orient);
  output_slide(slidenum,image,xdim,bestbbox,resolution);
}

int size_check(int *bbox, int resolution, double *area, int num, int orientation)
{ /* returns true if slide satisfies minimum size constraints.
     Returns area of slide in *area, regardless of constraints.
     if num != 0, output warning message if small */
  int iwid, iht, retval;
  double wid, ht;
  double minwid[2], minht[2];
  minwid[HORIZ] = 1.28;
  minwid[VERT] = 0.88;
  minht[HORIZ] = 0.88;
  minht[VERT] = 1.28;
  iwid = bbox[Xbot] - bbox[Xtop] + 1;
  iht = bbox[Ybot] - bbox[Ytop] + 1;
  wid = (double)iwid / resolution;
  ht = (double)iht / resolution;
  retval = 1;
  if (wid < minwid[orientation]) {
    retval = 0;
    if (num) fprintf(stderr,"Warning! slide %d wid %.3f < %.2f\n",num,wid,
			minwid[orientation]);
  }
  if (ht < minht[orientation]) {
    retval = 0;
    if (num) fprintf(stderr,"Warning! slide %d ht %.3f < %.2f\n",num,ht,
			minht[orientation]);
  }
  *area = wid * ht;
  return retval;
}

void output_slide(int num,unsigned char *image,int xsize,int *bbox,int resolution)
{
  int iwid, iht;
  double wid, ht;
  char buf[80];
  FILE *fp;
  int i, k;

  iwid = bbox[Xbot] - bbox[Xtop] + 1;
  iht = bbox[Ybot] - bbox[Ytop] + 1;
  wid = (double)iwid / resolution;
  ht = (double)iht / resolution;
  TRACE(1,(DF,"Slide %d %.2f x %.2f\n",num,wid,ht));
  TRACE(3,(DF,"Slide %d bbox: %d %d %d %d\n",num,bbox[0],bbox[1],bbox[2],bbox[3]));
  sprintf(buf,"%04d.pnm",num);
  fp = ftry(buf,"w");
  fprintf(fp,"P6\n%d %d\n255\n",iwid,iht);
  for (i=bbox[Ytop]; i<=bbox[Ybot]; i++) {
    k = (i * xsize + bbox[Xtop]) * 3;
    fwrite(&(image[k]),3,iwid,fp);
  }
  fclose(fp);
}

int find_non_black(unsigned char *image,int xdim,int ydim,int *bbox,int nrows,
		double exp_wid, double exp_ht, double slope)
{ /* given image of xdim by ydim, look in bbox region and return new bbox
     with a tighter fit to the non-black area.  image is detected at a point
     which maximimzes the avg difference between the nrows rows on either side.
     Slope is the rate at which the weighting factor (preferring edges near
     the expected value) falls off.  Returns false if it looks like there was
     no slide in the image (white image)
  */

  double *rowenergy, *colenergy;
  int i, j, indx;
  int new_start_y, new_fini_y;
  int new_start_x, new_fini_x;

  rowenergy = alloc(ydim * sizeof(double));
  colenergy = alloc(xdim * sizeof(double));
  TRACE(1,(DF,"find_non_black bbox is %d %d %d %d\n",bbox[0],bbox[1],bbox[2],bbox[3]));

  for (i=bbox[Ytop]; i<=bbox[Ybot]; i++) {
    rowenergy[i] = 0.0;
    for (j=bbox[Xtop]; j<=bbox[Xbot]; j++) {
	indx = (i * xdim + j) * 3;
	rowenergy[i] += image[indx] + image[indx+1] + image[indx+2];
    }
    rowenergy[i] /= bbox[Xbot] - bbox[Xtop] + 1;
  }
  if (rowenergy[bbox[Ytop]] > 600) {return 0;}

  for (j=bbox[Xtop]; j<=bbox[Xbot]; j++) {
    colenergy[j] = 0.0;
    for (i=bbox[Ytop]; i<=bbox[Ybot]; i++) {
	indx = (i * xdim + j) * 3;
	colenergy[j] += image[indx] + image[indx+1] + image[indx+2];
    }
    colenergy[j] /= bbox[Ybot] - bbox[Ytop] + 1;
  }
#ifdef DEBUG
  TRACE(3,(DF,"printing col energies from %d to %d\n",bbox[Xtop],bbox[Xbot]));
  for (i=bbox[Xtop]; i<=bbox[Xbot]; i++) {
    TRACE(3,(DF,"%4d %5.1f ",i,colenergy[i]));
    if ((i%8) == 7) {
	TRACE(3,(DF,"\n"));
    }
  }
  TRACE(3,(DF,"printing row energies from %d to %d\n",bbox[Ytop],bbox[Ybot]));
  for (i=bbox[Ytop]; i<=bbox[Ybot]; i++) {
    TRACE(3,(DF,"%4d %5.1f ",i,rowenergy[i]));
    if ((i%8) == 7) {
	TRACE(3,(DF,"\n"));
    }
  }
#endif
  new_start_y = find_step_increase(rowenergy,bbox[Ytop],bbox[Ybot],nrows,exp_ht,slope);
  TRACE(1,(DF,"got top y as %d\n",new_start_y));
  new_fini_y = find_step_increase(rowenergy,bbox[Ybot],bbox[Ytop],nrows,exp_ht,slope);
  TRACE(1,(DF,"got bot y as %d\n",new_fini_y));
  new_start_x = find_step_increase(colenergy,bbox[Xtop],bbox[Xbot],nrows,exp_wid,slope);
  TRACE(1,(DF,"got left x as %d\n",new_start_x));
  new_fini_x = find_step_increase(colenergy,bbox[Xbot],bbox[Xtop],nrows,exp_wid,slope);
  TRACE(1,(DF,"got right x as %d\n",new_fini_x));
  bbox[Ytop] = new_start_y;
  bbox[Ybot] = new_fini_y;
  bbox[Xtop] = new_start_x;
  bbox[Xbot] = new_fini_x;
}

int find_step_increase(double *energy,int start,int fini,int nrows,double exp_size,
	double slope)
{ /* given array energy, start at start and continue toward fini,
     stopping and returning the row which maximizes the difference between the
     avg of the next nrows rows and the avg of the previous nrows rows.
     exp_size is the expected size of the slide in this dimension, and is used
     to set up a weighting factor favoring differences near expected location.
     That is unity at the expected edge, and drops to zero in the expected
     slide center for slope 1.0.  If slope is different, falls off accordingly.
     Dropoff before expected edge (in background) has same rate */
  int inc; /* appropriate increment */
  int jump; /* nrows times the inc */
  double center, exp_edge, weight;
  int i, ii, jj, kk;
  double prevsum, nextsum;
  double prevavg, nextavg;
  double diff, bestdiff;
  int besti;

  inc = (start < fini) ? 1 : -1;
  TRACE(1,(DF,"find_step_increase(.,%d,%d)\n",start,fini));
  if (abs(start - fini) < (2*nrows)) {
    fprintf(stderr,"find_step_increase: need %d rows at least\n",2*nrows);
    return -1;
  }
  center = (fini + start) / 2.0;
  exp_edge = center - (exp_size/2.0) * inc;
  TRACE(3,(DF,"[%d,%d] center = %.1f, edge expected at %.1f\n",start,fini,center,exp_edge));
   /*	init prevsum, nextsum.  Don't need to find average, since they
	have the same number of samples, but nicer for debug code! */
  bestdiff = 0;
  prevsum = 0;
  kk = start;
  ii = kk;
  for (i=0; i<nrows; i++) {
    prevsum += energy[ii];
    ii += inc;
  }
  nextsum = 0;
  jj = ii;
  for (i=0; i<nrows; i++) {
    nextsum += energy[jj];
    jj += inc;
  }
   /* check difference */
  jump = nrows * inc;
  diff = 0;
  for (i=ii; i != (fini-(inc*nrows)); i+=inc) {
    weight = 1.0 - slope * fabs( i - exp_edge) / (exp_size/2);
#if 0
    if (weight > 0) weight *= weight; /* try square of simple linear weight */
#endif
    if (weight < 0.1) weight = 0.1;
    prevavg = prevsum / nrows;
    nextavg = nextsum / nrows;
    diff = (nextavg - prevavg) * weight;
    TRACE(5,(DF,"%d %.1f ^ %.1f (%.1f) wt %.2f\n",i,prevavg,nextavg,diff,weight));
    if (diff > bestdiff) {
	besti = i;
	bestdiff = diff;
	TRACE(3,(DF,"%d %.1f ^ %.1f (%.1f) wt %.2f\n",i,prevavg,nextavg,diff,weight));
    }
/* kk pts to start of prev sum, i is middle, jj is end of next sum */
    prevsum -= energy[kk];
    kk += inc;
    prevsum += energy[i+inc];
    nextsum -= energy[i];
    nextsum += energy[jj+inc];
    jj += inc;
/* now ready to test for transition at ii+inc in next loop iteration*/
  }
  return besti;
}

char *getlin(FILE *fp, int len, char *buf)
{
  if (fgets(buf,len,fp) == NULL) {
    fprintf(stderr,"Error on read\n");
    exit(2);
  }
  if (buf[0] == '#') getlin(fp,len,buf);
  if (buf[0] == '\n') getlin(fp,len,buf);
}
